#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <setjmp.h>
#include "rally.h"

extern int move_fd;

char* rally_host = RALLY_DEFAULT_HOST;
int rally_port = RALLY_DEFAULT_PORT;

int board_size;
int expected_move_serial;
player *players;
int winner;
int player_number = -1;

static void info(char* s) {
	fprintf(stderr, "Server Message: %s\n", s);
}

static char* get_line() {
	extern jmp_buf env;
	static char buf[1024];
	static char *p = buf;

	char tmp[sizeof(buf)], *tmpp;
	int len;

	for(;;) {
		len = read(move_fd, tmp, 1);
		if (len == -1)
			bailsys("read() failed");

		if (len == 0) {
			log("I: Server closed connection.");
			close(move_fd);
			exit(0);
		}

		tmpp = tmp;
		do {
			if (*tmpp == '\r' || *tmpp == '\n') {
				*p = '\0';

				/* ignore empty messages or between \r\n */
				if (p != buf) {
					p = buf;
					/* if it's an INFO message, print it out */
					if (!strncmp(buf, "INFO ", 5))
						info(buf+5);
					else {
						if (!strcmp(buf, "ABORT"))
							longjmp(env, 1);
						debug("V: Incoming: %s", buf);
						return buf;
					}
				}

				*p = '\0';
				continue;
			}

			/* only write if we have space */
			if (p - buf < sizeof(buf)-1)
				*(p++) = *tmpp;
		} while (tmpp++,--len);
	}
	/* Should never get here */
	assert(0);
}

void am_watcher() {
	send_printf(move_fd, "WATCHER\n");
}

void set_name(char *name) {
	char *msg;

	send_printf(move_fd, "HELL1 %s\n", name);

	/* Wait for a welcome message */
	msg = get_line();
	if (strncmp(msg, "WELCOME ", 8)) {
		log("E: Bad response from server: %s", msg);
		exit(1);
	}
	player_number = atoi(msg+8);
}

static void parse_board(char *np) {
	char *line, *p;
	int r, c, i;
	assert(sscanf(np, "%d%d", &number_of_players, &board_size) == 2);
	assert(number_of_players >= 2);

	players = calloc(sizeof(player), MAX_CLIENTS);
	for (i = 0; i < MAX_CLIENTS; i++) {
		players[i].health = STARTING_HEALTH;
	}

	for (r = 0; r < board_size; r++) {
		line = get_line();
		p = line;
		for (c = 0; c < board_size; c++) {
			assert(sscanf(p, "%d", &board[r][c]) == 1);
			if (board[r][c] != -1) {
				/* FIXME bounds checking */
				players[board[r][c]].r = r;
				players[board[r][c]].c = c;
				debug("Got player %d at %d,%d", board[r][c], r, c);
			}
			while (*p != ' ' && *p != '\n' && *p != '\r') p++;
			while (*p == ' ' || *p == '\n' || *p == '\r') p++;
		}
	}
	for(;;) {
		char tmp[50];
		line = get_line();
		if (strncmp("PLAYER ", line, 7))
			break;
		sscanf(line+7, "%d %d %s",  &r, &c, tmp);
		strcpy(players[r].name, tmp);
		players[r].dir = c; /* FIXME bounds checking */
	}
	assert(strcmp(line, "THE END")==0);
	debug("Got a board of size %d with %d players", board_size, number_of_players);
}

void rally_wait_for_game() {
	char *msg;
	winner = -1;
	for(;;) {
		msg = get_line();
		debug("Got line %s", msg);
		if (!strcmp(msg, "START")) {
			expected_move_serial = 1;
			return;
		}
		if (!strncmp(msg, "BOARD ", 6))
			parse_board(msg+6);
	}
}

void wait_until_ready_to_move() {
	char *msg;
	for(;;) {
		msg = get_line();
		int n;
		if (!strncmp(msg, "GAME OVER ", 10)) {
			winner = atoi(msg+10);
			return;
		}
		if (strncmp(msg, "MOVE ", 5))
			continue;
		n = atoi(msg+5);
		assert(n == expected_move_serial);
		expected_move_serial++;
		break;
	}
}

void submit_moves(char *moves) {
	char copy_moves[MOVES_PER_ROUND+1];
	if (!valid_moves(moves))
		bail("You played an invalid move (%s).", moves);
	strncpy(copy_moves, moves, MOVES_PER_ROUND);
	copy_moves[MOVES_PER_ROUND] = '\0';
	send_printf(move_fd, "MOVE %s\n", copy_moves);
}

int rally_connect() {
	struct sockaddr_in si;
	struct hostent *hp;

	if (move_fd != -1)
		return -EBUSY;

	if ((hp = gethostbyname(rally_host)) == NULL)
		bailsys("Failed to resolve host");

	if (!hp->h_addr_list[0])
		bail("No address records for host %s.", rally_host);

	if ((move_fd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
		bailsys("socket() failed");

	memset(&si, 0, sizeof(si));
	si.sin_family = AF_INET;
	si.sin_port = htons(rally_port);
	memcpy(&si.sin_addr, *(hp->h_addr_list), hp->h_length);

	if ((connect(move_fd, (struct sockaddr*)&si, sizeof(si))) == -1)
		bailsys("connect() failed");

	if (players) {
		free(players);
		players = NULL;
	}

	return 0;
}

int rally_disconnect() {
	if (move_fd == -1)
		return -EBADF;

	close(move_fd);
	return 0;
}

void wait_for_results(void (*printfunc)(int)) {
	char *msg;
	int move_num;
	for(;;) {
		int n;
		char move, *p;
		msg = get_line();
		if (!strncmp(msg, "GAME OVER ", 10)) {
			winner = atoi(msg+10);
			return;
		}
		if (!strncmp(msg, "DEAD ", 5)) {
			int p, r, c;
			p = atoi(msg+5);
			r = players[p].r;
			c = players[p].c;
			players[p].health = 0;
			board[r][c] = -1;
			continue;
		}
		if (strncmp(msg, "DOMOVE ", 7))
			break;
		p = msg+7;
		assert(sscanf(p, "%d%d", &n, &move_num) == 2);
		while (*p && *p != ' ' && *p != '\n' && *p != '\r') p++;
		while (*p && (*p == ' ' || *p == '\n' || *p == '\r')) p++;
		while (*p && *p != ' ' && *p != '\n' && *p != '\r') p++;
		while (*p && (*p == ' ' || *p == '\n' || *p == '\r')) p++;
		move = *p;
		move_player(n, move, move_num, 2);
		if (printfunc)
			printfunc(move_num);
	}
}

