#define _BSD_SOURCE
#include <assert.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <unistd.h>
#include "rally.h"

#define GAMES_UNTIL_RESTART	25  /* Set to -1 to avoid restarting */
#define MAX_MOVES_PER_GAME	200 /* To avoid games going on forever */
#define SECONDS_BETWEEN_GAMES	60  /* Games will start automatically if they can */

int board_size = 10;
int max_serial = MAX_MOVES_PER_GAME;

static int listen_fd = -1;
static int num_clients = 0;
client *clients;
int num_players = 0, players_alive;
int board[MAX_BOARD_SIZE][MAX_BOARD_SIZE];
static int *has_moved, moves_waiting_on;
static int games_until_restart = GAMES_UNTIL_RESTART;
player *players;
int winner;
char *argv0;

static int game_in_progress = 0;
static int move_serial;
int expected_move_serial;
static struct timeval start_time;
static struct timeval last_game_time;

extern void print_board(int move_num);

static void print_winners(int winner) {
	int i = 0;
	FILE* resf = fopen("results.txt", "a+");
	if (!resf) {
		log("E: Unable to append to log file.");
		return;
	}
	switch (winner) {
		case -1: /* no winner? */
			fprintf(resf, "NoWinner.\n");
			break;
		case -2: /* all dead */
			fprintf(resf, "All died.\n");
			break;
		case -3: /* aborted */
			fprintf(resf, "Aborted.\n");
			break;
		case -4: /* tied dead */
			for (i = 0; i < MAX_CLIENTS; i++) {
				if (clients[i].state != STATE_READY)
					continue;
				fprintf(resf, "%s ", players[i].name);
			}
			fprintf(resf, "\n");
			break;
		default:
			fprintf(resf, "%s\n", players[winner].name);
	}
	fclose(resf);
}

static void end_game(int winner) {
	if (!game_in_progress)
		return;
	timerclear(&start_time);
	game_in_progress = 0;
	broadcast_printf("GAME OVER %d\n", winner);
	log("Game over. Winner is %d", winner);
	print_winners(winner);
	if (--games_until_restart == 0) {
		broadcast_printf("INFO Server restarting ...\n");
		int i;
		for (i = 0; i < MAX_CLIENTS; i++) {
			if (clients[i].fd == -1) 
				continue;
			shutdown(clients[i].fd, SHUT_WR);
			close(clients[i].fd);
		}
		shutdown(listen_fd, SHUT_WR);
		close(listen_fd);
		execlp(argv0, argv0, NULL);
	}
	gettimeofday(&last_game_time, NULL);
}

static void reset_timer() {
	gettimeofday(&start_time, NULL);
}

void play_out_moves() {
	int move_num, cn, shooting;
	debug("Playing out moves...");
	for (move_num = 1; move_num <= MOVES_PER_ROUND; move_num++) {
		for (shooting = 0; shooting <= 1; shooting++) {
			for (cn = 0; cn < MAX_CLIENTS; cn++) {
				if (clients[cn].state != STATE_READY)
					continue;
				move_player(cn, clients[cn].nextmove[move_num-1], move_num, shooting);
				if (shooting ^ (clients[cn].nextmove[move_num-1] != MOVE_SHOOT &&
						clients[cn].nextmove[move_num-1] != MOVE_PUNCH)) {
					broadcast_printf("DOMOVE %d %d %c\n", cn, move_num, clients[cn].nextmove[move_num-1]);
					/* print_board(0); */
				}
				if (players_alive == 0) {
					end_game(-2);
					return;
				}
				if (players_alive == 1) {
					end_game(winner);
					return;
				}
			}
		}
		usleep(50*1000);
	}
	broadcast_printf("ROUND OVER\n");
}

void call_for_moves() {
	int i;
	for (i = 0; i < MAX_CLIENTS; i++)
		if (clients[i].state == STATE_READY || clients[i].state == STATE_WATCHER)
			send_printf(clients[i].fd, "MOVE %d\n", move_serial);
	memset(has_moved, 0, sizeof(int)*MAX_CLIENTS);
	moves_waiting_on = players_alive;
	move_serial++;
	if (move_serial > max_serial) {
		end_game(-4);
	}
	reset_timer();
}

void place_players() {
	/* Finds an appropriate set of starting positions for the players, so
	 * that they are not firing at each other when they start.
	 */

	int i, r = 0, c = 0;
	direction next_heading = NORTH;

	for (i = 0; i < MAX_CLIENTS; i++) {
		if (clients[i].state != STATE_READY)
			continue;
		players[i].dir = next_heading;

		switch (next_heading) {
			case NORTH:
				r = board_size - 1;
				c = (board_size/2)-(i/4)-1;
				next_heading = SOUTH;
				break;
			case SOUTH:
				r = 0;
				c = (board_size/2)+(i/4);
				next_heading = EAST;
				break;
			case EAST:
				r = (board_size/2)-(i/4)-1;
				c = 0;
				next_heading = WEST;
				break;
			case WEST:
				r = (board_size/2)+(i/4);
				c = board_size - 1;
				next_heading = NORTH;
				break;
			default:
				assert(0);
		}
		players[i].r = r;
		players[i].c = c;
		players[i].health = STARTING_HEALTH;
		board[r][c] = i;
	}
}

void broadcast_board() {
	int i, j, r, c;
	log("I: Broadcasting boards");
	for (i = 0; i < MAX_CLIENTS; i++) {
		if (clients[i].fd == -1)
			continue;
		send_printf(clients[i].fd, "BOARD %d %d\n", num_players, board_size);
		for (r = 0; r < board_size; r++) {
			for (c = 0; c < board_size; c++) {
				send_printf(clients[i].fd, "%d ", board[r][c]);
			}
			send_printf(clients[i].fd, "\n");
		}
		/* and now player directions */
		for (j = 0; j < MAX_CLIENTS; j++) {
			if (clients[j].state != STATE_READY)
				continue;
			send_printf(clients[i].fd, "PLAYER %d %d %s\n", j, players[j].dir, players[j].name);
		}
		send_printf(clients[i].fd, "THE END\n");
	}
}

void begin_game() {
	int i;

	timerclear(&last_game_time);
	if (game_in_progress) {
		end_game(-3);
		return;
	}

	if (num_players < 2) {
		log("E: Not enough players to start a game.");
		return;
	}

	/* Revive all players from the dead */
	for (i = 0; i < MAX_CLIENTS; i++) {
		if (clients[i].state == STATE_DEAD)
			clients[i].state = STATE_READY;
	}

	game_in_progress = 1;
	move_serial = 1;
	players_alive = num_players;
	winner = -1;

	memset(board, -1, sizeof(board));
	place_players();
	broadcast_board();
	broadcast_printf("START\n");
	call_for_moves();
}

/**** Networking stuff ****/

static void bind_to_port() {
	struct sockaddr_in si;
	int val = 1;

	if ((listen_fd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
		bailsys("socket() failed");

	if (setsockopt(listen_fd, SOL_SOCKET, SO_REUSEADDR, &val, sizeof(val)) == -1)
		bailsys("setsockopt() failed");

	si.sin_family = AF_INET;
	si.sin_port = htons(RALLY_DEFAULT_PORT);
	si.sin_addr.s_addr = INADDR_ANY;

	if (bind(listen_fd, (struct sockaddr*)&si, sizeof(si)) == -1)
		bailsys("bind() failed");

	if (listen(listen_fd, 20) == -1)
		bailsys("listen() failed");
}

static void close_port() {
	shutdown(listen_fd, SHUT_RDWR);
	close(listen_fd);
}

static int accept_new() {
	int new_client;
	int i;
	struct sockaddr sa;
	int sa_size = sizeof(sa);
	if ((new_client = accept(listen_fd, &sa, &sa_size)) == -1)
		bailsys("accept() failed");
	if (num_clients == MAX_CLIENTS) {
		const char *msg = "E: too many clients\n";
out:
		write(new_client, msg, sizeof(msg));
		close(new_client);
		log("E: Refused connection from %s",
				inet_ntoa(((struct sockaddr_in*)&sa)->sin_addr));
		return 0;
	}
	/* find a vacant spot */
	for (i = 0; i < MAX_CLIENTS; i++)
		if (clients[i].fd == -1)
			break;
	if (i == MAX_CLIENTS) {
		goto out;
	}
	players[i].name[0] = '\0';
	clients[i].fd = new_client;
	clients[i].state = STATE_UNINIT;
	clients[i].msgp = clients[i].msg;
	log("I: Accepted connection from %s (now client %d)",
			inet_ntoa(((struct sockaddr_in*)&sa)->sin_addr), i);
	if (game_in_progress) {
		clients[i].state = STATE_DEAD;
		send_printf(clients[i].fd, "INFO Game is in progress. Please wait for the next game.\n");
	}
	return 1;
}

static void make_it_all_happen() {
	play_out_moves();
	/* print_board(0); */
	if (players_alive > 1)
		call_for_moves();
}

static void do_message(int cn) {
	debug("I: Message from %d is %s", cn, clients[cn].msg);
	/* HELLO message (used for welcoming, and changing names) */
	if (!strncmp(clients[cn].msg, "HELL1 ", 6)) {
		/* Update their name */
		if (!clients[cn].msg[7]) {
			send_printf(clients[cn].fd, "INFO Missing player name.\n");
		}
		strncpy(players[cn].name, clients[cn].msg+6,
			sizeof(players[cn].name));
		players[cn].name[sizeof(players[cn].name)-1] = '\0';

		/* Send back ack */
		send_printf(clients[cn].fd, "WELCOME %d %s\n", cn,
				players[cn].name);

		/* Note that they're hopefully ready now. */
		if (clients[cn].state == STATE_UNINIT) {
			if (num_players < MAX_PLAYERS) {
				clients[cn].state = STATE_READY;
				num_players++;
				if (num_players == 2) {
#if 0
					/* begin_game(); */
#else
					printf("Press <ENTER> when ready to start a game, otherwise a game will start soon.\n");
					gettimeofday(&last_game_time, NULL);
#endif
				}
			} else {
				send_printf(clients[cn].fd,
						"INFO Too many players.\n");
				clients[cn].state = STATE_WATCHER;
			}
		}
	} else if (!strncmp(clients[cn].msg, "WATCHER", 7)) {
		clients[cn].state = STATE_WATCHER;
	} else if (!strncmp(clients[cn].msg, "MOVE ", 5)) {
		if (has_moved[cn]) {
			send_printf(clients[cn].fd, "INFO Duplicate MOVE command ignored.\n");
		} else if (strlen(clients[cn].msg+5) != MOVES_PER_ROUND) {
			send_printf(clients[cn].fd, "INFO Short MOVE command ignored.\n");
		} else {
			if (clients[cn].state != STATE_READY) {
				send_printf(clients[cn].fd, "INFO MOVE command ignored. You are dead.\n");
				return;
			}
			if (!valid_moves(clients[cn].msg+5)) {
				send_printf(clients[cn].fd, "INFO Invalid MOVE command ignored.\n");
				return;
			}
			has_moved[cn] = 1;
			memcpy(clients[cn].nextmove, clients[cn].msg+5, MOVES_PER_ROUND);
			moves_waiting_on--;
			if (moves_waiting_on <= 0)
				make_it_all_happen();
		}
	}
}

static void handle_client(int cn) {
	char buf[MSG_LEN+2];
	char *p;
	int len;
	len = read(clients[cn].fd, buf, MSG_LEN+2);
	if (len == -1) {
		if (errno == ECONNRESET)
			len = 0; /* let the next case deal with it */
		else
			bailsys("read() from client %d failed", cn);
	}

	/* Deal with closed sockets */
	if (len == 0) {
		shutdown(clients[cn].fd, SHUT_RDWR);
		close(clients[cn].fd);
		log("I: Connection to player %d (%s) closed.", cn, players[cn].name);
		clients[cn].fd = -1;
		num_clients--;
		if (clients[cn].state >= STATE_READY) {
			num_players--;
			if (clients[cn].state == STATE_READY) {
				players_alive--;
				moves_waiting_on--;
				board[players[cn].r][players[cn].c] = -1;
			}
		}
		clients[cn].state = STATE_UNINIT;
		end_game(-4);
		return;
	}

	p = buf;
	do {
		if (*p == '\r' || *p == '\n') {
			*clients[cn].msgp = '\0';

			/* ignore empty messages or between \r\n */
			if (clients[cn].msgp != clients[cn].msg)
				do_message(cn);

			clients[cn].msgp = clients[cn].msg;
			*clients[cn].msgp = '\0';
			continue;
		}

		/* only write if we have space */
		if (clients[cn].msgp - clients[cn].msg < MSG_LEN)
			*(clients[cn].msgp++) = *p;

	} while(p++,--len);
}

int main(int argc, char** argv) {
	int i;
	argv0 = argv[0];

	/* ignore SIGPIPE */
	signal(SIGPIPE, SIG_IGN);

	bind_to_port();

	clients = (client*)calloc(sizeof(client), MAX_CLIENTS);
	players = (player*)calloc(sizeof(player), MAX_CLIENTS);
	has_moved = (int*)malloc(sizeof(int)*MAX_CLIENTS);
	for (i = 0; i < MAX_CLIENTS; i++)
		clients[i].fd = -1;

	printf("Listening on port %d...\n", RALLY_DEFAULT_PORT);

	timerclear(&last_game_time);
	for(;;) {
		int max_fd;
		fd_set rfds;
		FD_ZERO(&rfds);

		FD_SET(0, &rfds);
		FD_SET(listen_fd, &rfds);
		max_fd = listen_fd;
		for (i = 0; i < MAX_CLIENTS; i++) {
			if (clients[i].fd != -1) {
				FD_SET(clients[i].fd, &rfds);
				if (clients[i].fd > max_fd)
					max_fd = clients[i].fd;
			}
		}

		struct timeval tv;
		struct timeval *use_tv = NULL;
		if (timerisset(&start_time)) {
			gettimeofday(&tv, NULL);
			tv.tv_sec += 1;
			tv.tv_sec -= start_time.tv_sec;
			tv.tv_usec -= start_time.tv_usec;
			if (tv.tv_usec < 0) {
				tv.tv_usec += 1000000;
				tv.tv_sec -= 1;
			}
			use_tv = &tv;
		} else if (!game_in_progress && timerisset(&last_game_time)) {
			gettimeofday(&tv, NULL);
			tv.tv_sec += SECONDS_BETWEEN_GAMES;
			tv.tv_sec -= last_game_time.tv_sec;
			tv.tv_usec -= last_game_time.tv_usec;
			if (tv.tv_usec < 0) {
				tv.tv_usec += 1000000;
				tv.tv_sec -= 1;
			}
			use_tv = &tv;
		}
		if ((i = select(max_fd+1, &rfds, NULL, NULL, use_tv)) == -1)
			bailsys("select() failed");

		if (i == 0) {
		    if (game_in_progress) {
			log("I: %d player(s) are too slow. Proceeding anyway.", moves_waiting_on);
			make_it_all_happen();
			continue;
		    } else
			begin_game();
		}
		if (FD_ISSET(0, &rfds)) {
			char buf[2];
			if (read(0, buf, 1) == 1 && buf[0] == '\n')
				begin_game();
		}

		if (FD_ISSET(listen_fd, &rfds))
			accept_new();

		for (i = 0; i < MAX_CLIENTS; i++) {
			if (clients[i].fd == -1 || !FD_ISSET(clients[i].fd, &rfds))
				continue;
			handle_client(i);
		}
	}

	close_port();
	return 0;
}
